from dataclasses import dataclass
import threading
import queue
import time
import asyncio
import socketserver
import logging
from typing import Iterator, Optional, Tuple, Set

import bleak
import click
import prometheus_client

from janet import Device, Channel

logger = logging.getLogger('janet.heyo.blebridge')


@dataclass
class State:
    client_count: int
    sent_bytes: int
    recv_bytes: int


class Client(socketserver.BaseRequestHandler):
    """Handles a single socket client."""
    def setup(self):
        self._tx = queue.Queue(100)
        self.server.register(self)

    def handle(self):
        def sender():
            while True:
                got = self._tx.get()
                if not got:
                    break
                self.request.send(got)

        sender = threading.Thread(target=sender)
        sender.start()
        try:
            while True:
                got = self.request.recv(16)
                if not got:
                    break
                self.server.put(got)
        finally:
            self._tx.put(None)
            sender.join()

    def finish(self):
        self.server.unregister(self)

    def put(self, data: bytes):
        self._tx.put_nowait(data)


class Bridge(socketserver.ThreadingTCPServer):
    """Bridges between a BLE based serial adapter and client sockets."""
    def __init__(self,
                 server_address: Tuple[str, int],
                 mac: str,
                 adapter: Optional[str] = None,
                 write_uuid: str = '0000fff2-0000-1000-8000-00805f9b34fb',
                 read_uuid: str = '0000fff1-0000-1000-8000-00805f9b34fb'):

        super().__init__(server_address, Client)
        self._mac = mac
        self._adapter = adapter
        self._write_uuid = write_uuid
        self._read_uuid = read_uuid

        self._lock = threading.RLock()
        self._clients: Set[Client] = set()
        self._tx: 'queue.Queue[bytes]' = queue.Queue(100)

        self._connected = False
        self._sent_bytes = 0
        self._recv_bytes = 0

        self._listeners = Channel()

    def listen(self, listener) -> None:
        self._listeners.listen(listener)

    def register(self, client: Client):
        with self._lock:
            self._clients.add(client)

    def unregister(self, client: Client):
        with self._lock:
            self._clients.remove(client)

    def _emit(self):
        device = Device(name=f'blebridge {self._mac}',
                        identifiers=[f'blebridge_{self._mac}'],
                        available=self._connected,
                        kind='blebridge')

        status = State(
            sent_bytes=self._sent_bytes,
            recv_bytes=self._recv_bytes,
            client_count=len(self._clients),
        )
        self._listeners.put(device, status)

    def run(self):
        server = threading.Thread(target=self.serve_forever)
        server.start()
        try:
            self.ble_run()
        finally:
            self.shutdown()
            server.join()

    def ble_run(self):
        self._emit()

        def recv(n, data: bytes):
            with self._lock:
                for client in self._clients:
                    client.put(data)
                self._recv_bytes += len(data)
                self._emit()

        async def run(address: str):
            async with bleak.BleakClient(address,
                                         adapter=self._adapter) as client:
                await client.start_notify(self._read_uuid, recv)
                try:
                    self._connected = True
                    self._emit()
                    while True:
                        await asyncio.sleep(0.2)
                        try:
                            tx = self._tx.get_nowait()
                            await client.write_gatt_char(
                                self._write_uuid, bytearray(tx))
                            self._sent_bytes += len(tx)
                            self._emit()
                        except queue.Empty:
                            pass
                finally:
                    await client.stop_notify(self._read_uuid)
                    self._connected = False
                    self._emit()

        while True:
            try:
                asyncio.run(run(self._mac))
            except (bleak.BleakError, AttributeError) as ex:
                logger.warning(f'Got an error "{ex}", retrying')
                time.sleep(5)

    def put(self, data: bytes):
        self._tx.put(data)


@click.command()
@click.option('--mac',
              type=str,
              required=True,
              help='MAC address to connect to.')
@click.option('--adapter',
              type=str,
              default='hci0',
              help='Adapter to cnnect via.')
@click.option('--port',
              type=int,
              default=7010,
              help='Port number to listen on.')
@click.option('--prometheus_port',
              type=int,
              default=7011,
              help='Port number to listen for monitoring on.')
def main(mac: str, adapter: str, port: int, prometheus_port: int):
    logging.basicConfig(
        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
        level=logging.INFO)

    import janet.prometheus

    prometheus_client.start_http_server(prometheus_port)
    prom = janet.prometheus.Client()

    addr = ('', port)
    bridge = Bridge(addr, mac, adapter)
    bridge.listen(prom.publish)

    bridge.run()


if __name__ == '__main__':
    main()
