import threading
import time
from dataclasses import dataclass
from typing import Iterator, Optional
import logging
import struct

from janet import Device

import serial
import libscrc
import prometheus_client

from . import schema

logger = logging.getLogger('heyo.blebridge')
logging.basicConfig(
    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
    level=logging.INFO)

_SOF = 0xD2


@dataclass
class Frame:
    kind: int
    payload: bytes


def framer(kind: int, payload: bytes) -> bytes:
    """Encodes the request into a frame.

    Note that the protocol is not symmetrical, so a frame generated by `framer` can't be decoded by `deframer`.
    """
    header = bytearray((_SOF, kind))
    crc = libscrc.modbus(header + payload)
    return header + payload + struct.pack('<H', crc)


def deframer(port: serial.Serial) -> Iterator[Frame]:
    def reader(port: serial.Serial) -> Iterator[int]:
        while True:
            for ch in port.read():
                yield ch

    src = reader(port)
    while True:
        while next(src) != _SOF:
            pass

        kind = next(src)
        length = next(src)

        payload = bytearray()
        while len(payload) != length:
            payload.append(next(src))

        crch, crcl = next(src), next(src)
        if libscrc.modbus(
                bytes((_SOF, kind, length)) + payload +
                bytes((crch, crcl))) != 0:
            logger.error('CRC mismatch, dropping frame')
            continue

        yield Frame(kind, bytes(payload))


def main():
    import janet.prometheus

    host = 'socket://192.168.1.4:9876'
    port = serial.serial_for_url(host)
    device = Device(identifiers=[f'heyo_{host}'], kind='heyo', name='heyo')

    prometheus_client.start_http_server(9874)
    prom = janet.prometheus.Client()

    def poker():
        while True:
            port.write(framer(3, b'\x00\x00\x00\x3e'))
            time.sleep(1)

    threading.Thread(target=poker, daemon=True).start()

    for frame in deframer(port):
        frame.device = device
        prom.publish(frame)
        msg = schema.decode(frame.payload)
        if msg:
            msg.device = device
            prom.publish(msg)


#        print(schema.decode(frame.payload))

if __name__ == '__main__':
    main()
