from dataclasses import dataclass
from typing import Sequence, Optional
import enum
import struct

from janet import Device

V = float
A = float
W = float


class Length(enum.IntEnum):
    STATUS = 124


@dataclass
class Status:
    """High level status of the BMS."""
    # Voltage of each cell.
    cells: Sequence[V]
    # Total voltage of the batter.
    sum: V
    # Discharge or charge current.
    current: A
    # Overall charge capacity as a ratio.
    capacity: float
    # Voltage of the most charged cell.
    maximum: V
    # Voltage of the least charged cell.
    minimum: V
    # Average voltage across all cells.
    average: V
    # Difference between the most and least charged cell.
    difference: V
    # Absolute power being supplied or charging with.
    power: W
    charge_enabled: bool
    discharge_enabled: bool

    others: Sequence[int]


def decode(payload: bytes) -> Optional[Status]:
    if len(payload) != Length.STATUS:
        return None

    mV = 1e-3

    fields = [
        struct.unpack_from('>H', payload, i)[0]
        for i in range(0, len(payload), 2)
    ]
    cells = fields[:32]
    try:
        cells = cells[:cells.index(0)]
    except ValueError:
        pass

    return Status(
        cells=tuple(cell * mV for cell in cells),
        sum=fields[40] / 10,
        current=(fields[41] - 30000) / 10,
        capacity=fields[42] / 1000,
        maximum=fields[43] * mV,
        minimum=fields[44] * mV,
        # balancing? 47.
        # fault? 52
        charge_enabled=fields[53],
        discharge_enabled=fields[54],
        average=fields[55] * mV,
        difference=fields[56] * mV,
        power=fields[57],
        others=fields[32:],
    )