import hashlib
import json
import dataclasses
import logging
from typing import Any, Sequence, Optional

import paho.mqtt.client as mqtt

from . import Device

logger = logging.getLogger('janet.mqtt')


def _uid(entity, device: Device) -> str:
    cls = entity.__class__
    kind = device.kind if device else None
    kind = kind or cls.__module__
    uid = hashlib.sha1(str(device.identifiers).encode()).hexdigest()[:13]

    if entity == device:
        return f'{kind}_{uid}'
    else:
        return f'{kind}_{uid}/{cls.__name__}'.lower().replace('.', '_')


class Client:
    def __init__(self, host: str, port: int = 1883):
        self._client = mqtt.Client()
        self._client.connect_async(host, port, 60)
        self._client.loop_start()

    def publish(self, device: Device, entity: Any):
        uid = _uid(entity, device)

        self._client.publish(
            f'test/{uid}/state',
            json.dumps(dataclasses.asdict(entity), default=str))

        available = getattr(device, 'available', None)
        if available is not None:
            self._client.publish(f'test/{uid}/power',
                                 'ON' if available else 'OFF',
                                 retain=True)

        dev = {
            'identifiers': device.identifiers,
            'name': device.name,
        }
        if device.manufacturer:
            dev['manufacturer'] = device.manufacturer

        for field in dataclasses.fields(entity):
            value = getattr(entity, field.name)
            if value is None:
                continue

            if field.type in (bool, Optional[bool]):
                topic = f'test/{uid}/{field.name}'
                self._client.publish(topic, 'ON' if value else 'OFF')
                self._client.publish(
                    f'homeassistant/binary_sensor/{uid}_{field.name}/config',
                    json.dumps({
                        'name': uid,
                        'state_topic': topic,
                        'unique_id': f'{uid}_{field.name}',
                        'device': dev,
                    }))
            elif field.type in (int, float):
                self._client.publish(f'test/{uid}/{field.name}',
                                     round(value, 3))
            elif field.type in (str, ):
                self._client.publish(f'test/{uid}/{field.name}', value)
            elif field.type in (Sequence[float], Sequence[int], bytes):
                pass
            elif field.type in (Optional[Device], Device):
                pass
            else:
                logger.info(f'Ignoring {field.type}')
