import abc
import dataclasses
import enum
import logging
from typing import Dict, Any, Sequence, Optional

import prometheus_client
import pint

from . import Device

logger = logging.getLogger('janet.mqtt')

Instance = Any

_LABELS = ('identifier', )


class Metric(abc.ABC):
    pass


class Gauge(Metric):
    def __init__(self, name, field):
        self.metric = prometheus_client.Gauge(name,
                                              'No description',
                                              labelnames=_LABELS)

    def publish(self, identifier, value) -> None:
        if value is not None:
            self.metric.labels(identifier).set(round(value, 6))


class Bytes(Metric):
    def __init__(self, name, field):
        self.metric = prometheus_client.Gauge(f'{name}_bytes',
                                              'No description',
                                              labelnames=_LABELS)

    def publish(self, identifier, value) -> None:
        self.metric.labels(identifier).inc(len(value))


class MultiGauge(Metric):
    def __init__(self, name, field):
        self.metric = prometheus_client.Gauge(name,
                                              'No description',
                                              labelnames=(
                                                  'identifier',
                                                  'index',
                                              ))

    def publish(self, identifier, values) -> None:
        for i, value in enumerate(values):
            self.metric.labels(identifier, str(i)).set(round(value, 6))


class Quantity(Metric):
    def __init__(self, name, field):
        self.metric = prometheus_client.Gauge(name,
                                              'No description',
                                              labelnames=_LABELS)

    def publish(self, identifier, value) -> None:
        if value is not None:
            self.metric.labels(identifier).set(value.m)


class Info(Metric):
    def __init__(self, name, field):
        self.metric = prometheus_client.Info(name,
                                             'No description',
                                             labelnames=_LABELS)

    def publish(self, identifier, value) -> None:
        if value:
            self.metric.labels(identifier).info({'value': value})


class Enum(Metric):
    def __init__(self, name, field):
        states = [x.name.lower() for x in field.type]
        self.metric = prometheus_client.Enum(name,
                                             'No description',
                                             states=states)

    def publish(self, value: enum.Enum) -> None:
        self.metric.state(value.name.lower())


def _issubclass(typ, cls) -> bool:
    try:
        return issubclass(typ, cls)
    except TypeError:
        return False


def metric_factory(name: str, field: dataclasses.Field) -> Optional[Metric]:
    if isinstance(field.type, pint.Unit):
        return Quantity(name, field)
    elif field.type in (int, float, bool, Optional[int], Optional[float],
                        Optional[bool]):
        return Gauge(name, field)
    elif field.type in (Sequence[float], Sequence[int]):
        return MultiGauge(name, field)
    elif field.type in (bytes, ):
        return Bytes(name, field)
    elif field.type in (str, Optional[str]):
        return Info(name, field)
    elif field.type in (Device, Optional[Device]):
        return None
    elif _issubclass(field.type, enum.IntEnum):
        return Enum(name, field)
    else:
        logger.warning(f'Dropping {name} due to unsupported type {field.type}')
        return None


class Client:
    def __init__(self):
        self._metrics = {}

    def publish(self, device: Device, entity: Any):
        cls = entity.__class__
        kind = device.kind or cls.__module__
        name = f'{kind}_{cls.__name__}'.lower().replace('.', '_')
        identifier = device.identifiers[0]

        for field in dataclasses.fields(entity):
            fqn = (name, field.name)
            if fqn not in self._metrics:
                self._metrics[fqn] = metric_factory(f'{name}_{field.name}',
                                                    field)

            metric = self._metrics.get(fqn)
            if metric:
                metric.publish(identifier, getattr(entity, field.name))
